from collections.abc import Mapping

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.db import transaction
from rest_framework.serializers import Serializer


User = get_user_model()


class ObjectPermissionsAssignmentMixin(Serializer):
    """
    A serializer mixin that provides an easy way to assign permissions
    to given users and/or group when an object is created or updated.
    """

    def save(self, **kwargs):
        created = self.instance is not None

        result = super().save(**kwargs)

        permissions_map = self.get_permissions_map(created)
        self.assign_permissions(permissions_map)

        return result

    def get_permissions_map(self, created):
        """
        Return a map where keys are permissions
        and values are list of users and/or groups
        """
        raise NotImplementedError

    def assign_permissions(self, permissions_map):
        """
        Assign the permissions to their associated users/groups.
        """
        # Import at runtime (see: DjangoObjectPermissionsFilter.filter_queryset)
        from guardian.shortcuts import assign_perm

        assert isinstance(permissions_map, Mapping), (
            'Expected %s.get_permissions_map to return a dict, got %s instead.'
            % (self.__class__.__name__, type(permissions_map).__name__)
        )

        with transaction.atomic():
            for permission, assignees in permissions_map.items():
                users = [u for u in assignees if isinstance(u, User)]
                groups = [g for g in assignees if isinstance(g, Group)]

                # TODO: support Django Guardian bulk permission assigning
                # Currently, trying to assign a permission to multiple
                # users or groups can result uniqueness constraint error
                # in database level due to `assign_perm` not checking
                # the existance of a permission before inserting
                for user in users:
                    assign_perm(permission, user, self.instance)
                for group in groups:
                    assign_perm(permission, group, self.instance)



from collections import OrderedDict

from django.contrib.auth.models import Group, User
from django.test import TestCase
from rest_framework import serializers
from rest_framework.test import APIRequestFactory

from rest_framework_guardian.serializers import ObjectPermissionsAssignmentMixin
from tests.models import BasicModel


factory = APIRequestFactory()


class BasicSerializer(ObjectPermissionsAssignmentMixin, serializers.ModelSerializer):
    class Meta:
        model = BasicModel
        fields = '__all__'

    def get_permissions_map(self, created):
        current_user = self.context['request'].user
        readers = Group.objects.get(name='readers')

        return {
            'view_%s' % BasicModel._meta.model_name: [current_user, readers],
            'change_%s' % BasicModel._meta.model_name: [current_user],
        }


class ObjectPermissionsAssignmentIntegrationTests(TestCase):
    """
    Integration tests for the object level permissions assignment API.
    """

    def setUp(self):
        create = User.objects.create_user
        self.writer = create('writer', password='password')
        self.reader = create('reader', password='password')
        self.no_perms = create('no_perms', password='password')

        # Add readers to our reader group
        reader_group = Group.objects.create(name='readers')
        reader_group.user_set.add(self.reader)

    def create_object(self):
        request = factory.post('/')
        request.user = self.writer

        serializer = BasicSerializer(
            data={'text': 'test'},
            context={'request': request},
        )
        serializer.is_valid(raise_exception=True)
        instance = serializer.save()

        return instance

    def test_read_permissions(self):
        instance = self.create_object()

        # Reader is indirectly assigned via its group permissions
        self.assertTrue(self.writer.has_perm('view_basicmodel', instance))
        self.assertTrue(self.reader.has_perm('view_basicmodel', instance))
        self.assertFalse(self.no_perms.has_perm('view_basicmodel', instance))

    def test_change_permissions(self):
        instance = self.create_object()

        self.assertTrue(self.writer.has_perm('change_basicmodel', instance))
        self.assertFalse(self.reader.has_perm('change_basicmodel', instance))
        self.assertFalse(self.no_perms.has_perm('change_basicmodel', instance))

    def test_delete_permissions(self):
        instance = self.create_object()

        # No user should be assigned delete permissions
        self.assertFalse(self.writer.has_perm('delete_basicmodel', instance))
        self.assertFalse(self.reader.has_perm('delete_basicmodel', instance))
        self.assertFalse(self.no_perms.has_perm('delete_basicmodel', instance))


class ObjectPermissionsAssignmentImplementationTests(TestCase):

    def test_get_permissions_map_should_return_a_mapping(self):
        for return_value in [dict(), OrderedDict()]:
            class TestSerializer(BasicSerializer):
                def get_permissions_map(self, created):
                    return return_value  # noqa: B023

            serializer = TestSerializer(data={'text': 'test'})
            serializer.is_valid(raise_exception=True)
            self.assertIsInstance(serializer.save(), BasicModel)

    def test_get_permissions_map_error_message(self):
        error_message = (
            'Expected InvalidSerializer.get_permissions_map '
            'to return a dict, got list instead.'
        )

        class InvalidSerializer(BasicSerializer):
            def get_permissions_map(self, created):
                return []

        serializer = InvalidSerializer(data={'text': 'test'})
        serializer.is_valid(raise_exception=True)

        with self.assertRaisesMessage(AssertionError, error_message):
            serializer.save()



import warnings

from django.contrib.auth.models import Group, User
from django.test import TestCase
from rest_framework import authentication, generics, permissions, serializers, status
from rest_framework.test import APIRequestFactory

from rest_framework_guardian.filters import (
    DjangoObjectPermissionsFilter, ObjectPermissionsFilter)
from tests.models import BasicModel, BasicPermModel
from tests.utils import basic_auth_header


factory = APIRequestFactory()


class BasicSerializer(serializers.ModelSerializer):
    class Meta:
        model = BasicModel
        fields = '__all__'


class BasicPermSerializer(serializers.ModelSerializer):
    class Meta:
        model = BasicPermModel
        fields = '__all__'


# Custom object-level permission, that includes 'view' permissions
class ViewObjectPermissions(permissions.DjangoObjectPermissions):
    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': ['%(app_label)s.view_%(model_name)s'],
        'HEAD': ['%(app_label)s.view_%(model_name)s'],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }


class ObjectPermissionInstanceView(generics.RetrieveUpdateDestroyAPIView):
    queryset = BasicPermModel.objects.all()
    serializer_class = BasicPermSerializer
    authentication_classes = [authentication.BasicAuthentication]
    permission_classes = [ViewObjectPermissions]


object_permissions_view = ObjectPermissionInstanceView.as_view()


class ObjectPermissionListView(generics.ListAPIView):
    queryset = BasicPermModel.objects.all()
    serializer_class = BasicPermSerializer
    authentication_classes = [authentication.BasicAuthentication]
    permission_classes = [ViewObjectPermissions]
    filter_backends = [ObjectPermissionsFilter]


object_permissions_list_view = ObjectPermissionListView.as_view()


class GetQuerysetObjectPermissionInstanceView(generics.RetrieveUpdateDestroyAPIView):
    serializer_class = BasicPermSerializer
    authentication_classes = [authentication.BasicAuthentication]
    permission_classes = [ViewObjectPermissions]

    def get_queryset(self):
        return BasicPermModel.objects.all()


get_queryset_object_permissions_view = GetQuerysetObjectPermissionInstanceView.as_view()


class ObjectPermissionsIntegrationTests(TestCase):
    """
    Integration tests for the object level permissions API.
    """

    def setUp(self):
        from guardian.shortcuts import assign_perm

        # create users
        create = User.objects.create_user
        users = {
            'fullaccess': create('fullaccess', 'fullaccess@example.com', 'password'),
            'readonly': create('readonly', 'readonly@example.com', 'password'),
            'writeonly': create('writeonly', 'writeonly@example.com', 'password'),
            'deleteonly': create('deleteonly', 'deleteonly@example.com', 'password'),
        }

        # give everyone model level permissions, as we are not testing those
        everyone = Group.objects.create(name='everyone')
        model_name = BasicPermModel._meta.model_name
        app_label = BasicPermModel._meta.app_label
        perms = {
            'view': '{0}_{1}'.format('view', model_name),
            'change': '{0}_{1}'.format('change', model_name),
            'delete': '{0}_{1}'.format('delete', model_name),
        }
        for perm in perms.values():
            perm = '{0}.{1}'.format(app_label, perm)
            assign_perm(perm, everyone)
        everyone.user_set.add(*users.values())

        # appropriate object level permissions
        readers = Group.objects.create(name='readers')
        writers = Group.objects.create(name='writers')
        deleters = Group.objects.create(name='deleters')

        model = BasicPermModel.objects.create(text='foo')

        assign_perm(perms['view'], readers, model)
        assign_perm(perms['change'], writers, model)
        assign_perm(perms['delete'], deleters, model)

        readers.user_set.add(users['fullaccess'], users['readonly'])
        writers.user_set.add(users['fullaccess'], users['writeonly'])
        deleters.user_set.add(users['fullaccess'], users['deleteonly'])

        self.credentials = {}
        for user in users.values():
            self.credentials[user.username] = basic_auth_header(user.username, 'password')

    def test_can_read_list_permissions(self):
        request = factory.get('/', HTTP_AUTHORIZATION=self.credentials['readonly'])
        response = object_permissions_list_view(request)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data[0].get('id'), 1)

    def test_cannot_read_list_permissions(self):
        request = factory.get('/', HTTP_AUTHORIZATION=self.credentials['writeonly'])
        response = object_permissions_list_view(request)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.data, [])

    def test_deprecation_warning(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')

            DjangoObjectPermissionsFilter()

            self.assertEqual(len(w), 1)
            self.assertIs(w[0].category, DeprecationWarning)
            self.assertEqual(str(w[0].message),
                             '`DjangoObjectPermissionsFilter` has been renamed '
                             'to `ObjectPermissionsFilter` and will be removed '
                             'in the future.')
